@@ -14,13 +14,86 @@ import fnmatch
 import subprocess
 import functools
 import re
+import json
 
 import apt_pkg
 
 from UbuntuDrivers import kerneldetection
+from functools import cmp_to_key
 
 system_architecture = ''
 lookup_cache = {}
+custom_supported_gpus_json = '/etc/custom_supported_gpus.json'
+
+
+class NvidiaPkgNameInfo(object):
+    '''Class to process NVIDIA package names'''
+    def __init__(self, pkg_name):
+        self._pkg_name = pkg_name
+        self._obsolete_name_scheme = False
+        self._server = False
+        self._open = False
+        self._major_ver = -1
+        self._flavour = ''
+        self.is_valid = False
+        self._process_name(self._pkg_name)
+
+    def _process_name(self, name):
+        if 'nvidia' not in name:
+            logging.debug('NvidiaPkgNameInfo: %s is not an NVIDIA package. Skipping', name)
+            return
+
+        pattern = re.compile('nvidia-([0-9]+)')
+        match = pattern.match(name)
+
+        # Obsolete naming such as nvidia-340
+        if match:
+            self._obsolete_name_scheme = True
+            self._major_ver = int(match.group(1))
+            self._flavour = self._major_ver
+            self.is_valid = True
+            return
+
+        # Recent naming such as nvidia-driver-525
+        pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
+        match = pattern.match(name)
+
+        if match:
+            self._server = match.group(0).find('-server') != -1
+            self._open = match.group(0).find('-open') != -1
+            self._flavour = '%s%s%s' % (match.group(1),
+                                        '-server' if self._server else '',
+                                        '-open' if self._open else '')
+            self._major_ver = int(match.group(1))
+            self.is_valid = True
+
+        # Recent naming such as nvidia-headless[-no-dkms]-525 for gpgpu
+        pattern = re.compile('nvidia-headless(.*)-([0-9]+)(.*)')
+        match = pattern.match(name)
+
+        if match:
+            self._server = match.group(0).find('-server') != -1
+            self._open = match.group(0).find('-open') != -1
+            self._flavour = '%s%s%s' % (match.group(2),
+                                        '-server' if self._server else '',
+                                        '-open' if self._open else '')
+            self._major_ver = match.group(2)
+            self.is_valid = True
+
+    def has_obsolete_name_scheme(self):
+        return self._obsolete_name_scheme
+
+    def is_server(self):
+        return self._server
+
+    def is_open(self):
+        return self._open
+
+    def get_major_version(self):
+        return self._major_ver
+
+    def get_flavour(self):
+        return self._flavour
 
 
 def get_apt_arch():
@@ -190,6 +263,37 @@ def _apt_cache_modalias_map(apt_cache):
     return result2
 
 
+def path_get_custom_supported_gpus():
+    return custom_supported_gpus_json
+
+
+def package_get_nv_allowing_driver(did):
+    '''Get nvidia allowing driver for specific devices.
+
+    did: 0x1234
+    Return the situable nvidia driver version for it.
+    '''
+    path = path_get_custom_supported_gpus()
+    version = None
+    try:
+        with open(path, "r") as stream:
+            try:
+                gpus = list(json.load(stream)['chips'])
+                for gpu in gpus:
+                    if gpu['devid'] == did:
+                        version = gpu['branch'].split('.')[0]
+                        logging.info("Found a specific nv driver version %s for %s(%s)" %
+                                     (version, gpu['name'], did))
+                        break
+            except Exception:
+                logging.debug('package_get_nv_allowing_driver(): unexpected json detected.')
+                pass
+    except Exception:
+        logging.debug('package_get_nv_allowing_driver(): unable to read %s' % path)
+        pass
+    return version
+
+
 def packages_for_modalias(apt_cache, modalias):
     '''Search packages which match the given modalias.
 
@@ -205,8 +309,24 @@ def packages_for_modalias(apt_cache, modalias):
         packages_for_modalias.cache_maps[apt_cache_hash] = cache_map
 
     pat, bus_map = cache_map.get(modalias.split(':', 1)[0], (None, {}))
-    if pat is None or not pat.match(modalias):
-        return []
+    vid, did = _get_vendor_model_from_alias(modalias)
+    nvamd = None
+    found = 0
+    if vid == "10DE":
+        nvamd = package_get_nv_allowing_driver("0x" + did)
+        nvamdn = "nvidia-driver-%s" % nvamd
+        nvamda = "pci:v000010DEd0000%s*" % did
+        for p in apt_cache.packages:
+            if p.get_fullname().split(':')[0] == nvamdn:
+                bus_map[nvamda] = set([nvamdn])
+                found = 1
+        if nvamd is not None and not found:
+            logging.debug('%s is not in the package pool.' % nvamdn)
+
+    if not found:
+        if pat is None or not pat.match(modalias):
+            return []
+
     for alias in bus_map:
         if fnmatch.fnmatchcase(modalias.lower(), alias.lower()):
             for p in bus_map[alias]:
@@ -259,6 +379,27 @@ def _is_package_from_distro(apt_cache, pkg):
         return False
 
 
+def _pkg_get_open_preference(apt_cache, pkg):
+    '''Determine if -open package is prefered from apt Package object'''
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    records = apt_pkg.PackageRecords(apt_cache)
+    records.lookup(candidate.file_list[0])
+
+    try:
+        preference = records['Prefer-Variant']
+    except (KeyError, AttributeError):
+        logging.debug('_pkg_get_open_preference %s: package has Prefer-Open header, cannot determine', pkg.name)
+        return False
+
+    if preference not in ('Open', 'Closed'):
+        logging.debug('_pkg_get_open_preference %s: package has invalid Prefer-Open %s'
+                      'header, cannot determine preference', pkg.name, preference)
+        return None
+
+    return preference
+
+
 def _pkg_get_module(apt_cache, pkg):
     '''Determine module name from apt Package object'''
     depcache = apt_pkg.DepCache(apt_cache)
@@ -303,10 +444,46 @@ def _pkg_get_support(apt_cache, pkg):
     return support
 
 
-def _is_runtimepm_supported(pkg, alias):
+def _is_nv_allowing_runtimepm_supported(alias, ver):
+    '''alias: e.g. pci:v000010DEd000024BAsv0000103Csd000089C6bc03sc00i00'''
+    result = re.search('pci:v0000(.*)d0000(.*)sv(.*)', alias)
+    vid = result.group(1)
+    did = result.group(2)
+    if vid != "10DE":
+        return False
+    did = "0x%s" % did
+    path = path_get_custom_supported_gpus()
+    try:
+        with open(path, "r") as stream:
+            try:
+                gpus = list(json.load(stream)['chips'])
+                for gpu in gpus:
+                    if gpu['devid'] == did and 'runtimepm' in gpu['features']:
+                        if gpu['branch'].split('.')[0] != ver:
+                            logging.debug('Candidate version does not match %s != %s' %
+                                          (gpu['branch'].split('.')[0], ver))
+                            return False
+                        logging.info("Found runtimepm supports on %s." % did)
+                        return True
+            except Exception:
+                logging.debug('_is_nv_allowing_runtimepm_supported(): unexpected json detected')
+                pass
+    except Exception:
+        logging.debug('_is_nv_allowing_runtimepm_supported(): unable to read %s' % path)
+        pass
+    return False
+
+
+def _is_runtimepm_supported(apt_cache, pkg, alias):
     '''Check if the package supports runtimepm for the given modalias'''
     try:
-        m = pkg.candidate.record['PmAliases']
+        depcache = apt_pkg.DepCache(apt_cache)
+        candidate = depcache.get_candidate_ver(pkg)
+        records = apt_pkg.PackageRecords(apt_cache)
+        records.lookup(candidate.file_list[0])
+        section = apt_pkg.TagSection(records.record)
+        ver = candidate.ver_str.split('.')[0]
+        m = section['PmAliases']
     except (KeyError, AttributeError, UnicodeDecodeError):
         return False
     else:
@@ -315,6 +492,8 @@ def _is_runtimepm_supported(pkg, alias):
 
         n = m[m.find('(')+1: m.find(')')]
         modaliases = n.split(', ')
+        if _is_nv_allowing_runtimepm_supported(alias, ver):
+            return True
         return any(fnmatch.fnmatch(alias.lower(), regex.lower()) for regex in modaliases)
 
 
@@ -386,6 +565,24 @@ def _get_db_name(syspath, alias):
     return (vendor, model)
 
 
+def _is_open_prefered(apt_cache, pkg):
+    if not pkg.name.startswith('nvidia-'):
+        return False
+
+    preference = _pkg_get_open_preference(apt_cache, pkg)
+    if preference is None:
+        nvidia_info = NvidiaPkgNameInfo(pkg.name)
+        if nvidia_info.get_major_version() >= 560:
+            logging.debug('_is_open_prefered(%s): True', pkg.name)
+            return True
+    elif preference == 'Open':
+        logging.debug('_is_open_prefered(%s): True', pkg.name)
+        return True
+
+    logging.debug('_is_open_prefered(%s): False', pkg.name)
+    return False
+
+
 def set_nvidia_kms(value):
     '''Set KMS on or off for NVIDIA'''
     nvidia_kms_file = '/lib/modprobe.d/nvidia-kms.conf'
@@ -456,7 +653,8 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
                     'free': _is_package_free(apt_cache, p),
                     'from_distro': _is_package_from_distro(apt_cache, p),
                     'support': _pkg_get_support(apt_cache, p),
-                    'runtimepm': _is_runtimepm_supported(p, alias)
+                    'runtimepm': _is_runtimepm_supported(apt_cache, p, alias),
+                    'open_preferred': _is_open_prefered(apt_cache, p)
                 }
             (vendor, model) = _get_db_name(syspath, alias)
             if vendor is not None:
@@ -515,19 +713,16 @@ def _get_headless_no_dkms_metapackage(pkg, apt_cache):
     depcache = apt_pkg.DepCache(apt_cache)
     name = pkg.name
 
-    pattern = re.compile('nvidia-([0-9]+)')
-    match = pattern.match(name)
-    if match:
-        logging.debug('Legacy driver detected: %s. Skipping.', name)
+    nvidia_info = NvidiaPkgNameInfo(name)
+    if not nvidia_info.is_valid:
+        logging.debug('Unsupported driver detected: %s. Skipping' % name)
         return metapackage
 
-    pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
-    match = pattern.match(name)
-    if not match:
-        logging.debug('No flavour can be found in %s. Skipping.', name)
+    if nvidia_info.has_obsolete_name_scheme():
+        logging.debug('Legacy driver detected: %s. Skipping.' % name)
         return metapackage
 
-    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
+    candidate_flavour = nvidia_info.get_flavour()
     candidate = 'nvidia-headless-no-dkms-%s' % (candidate_flavour)
 
     try:
@@ -600,6 +795,7 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
                     'from_distro': _is_package_from_distro(apt_cache, p),
                     'recommended': True,
                     'support': _pkg_get_support(apt_cache, p),
+                    'open_preferred': _is_open_prefered(apt_cache, p),
                 }
 
     return packages
@@ -661,6 +857,7 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
                         'free': _is_package_free(apt_cache, p),
                         'from_distro': _is_package_from_distro(apt_cache, p),
                         'support': _pkg_get_support(apt_cache, p),
+                        'open_preferred': _is_open_prefered(apt_cache, p),
                     }
                 if vendor is not None:
                     packages[p.name]['vendor'] = vendor
@@ -779,12 +976,14 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     return result
 
 
-def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_oem=True, driver_string=''):
+def get_desktop_package_list(
+        apt_cache, sys_path=None, free_only=False, include_oem=True,
+        driver_string='', include_dkms=False):
     '''Return the list of packages that should be installed'''
     packages = system_driver_packages(
         apt_cache, sys_path, freeonly=free_only,
         include_oem=include_oem)
-    packages = auto_install_filter(packages, driver_string)
+    packages = auto_install_filter(packages, driver_string, get_recommended=False)
     if not packages:
         logging.debug('No drivers found for installation.')
         return packages
@@ -794,13 +993,15 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
 
     # ignore packages which are already installed
     to_install = []
-    for p in packages:
+    for p, _ in sorted(packages.items(),
+                       key=cmp_to_key(lambda left, right: _cmp_gfx_alternatives(left[0], right[0])),
+                       reverse=True):
         package_obj = apt_cache[p]
         if not package_obj.current_ver:
-            to_install.append(p)
 
             candidate = depcache.get_candidate_ver(package_obj)
             records.lookup(candidate.file_list[0])
+            to_install.append(p)
 
             # See if runtimepm is supported
             if records['runtimepm']:
@@ -817,7 +1018,11 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
             try:
                 modules_package = get_linux_modules_metapackage(apt_cache, p)
                 if modules_package and not apt_cache[modules_package].current_ver:
+                    if not include_dkms and "dkms" in modules_package:
+                        to_install.remove(p)
+                        continue
                     to_install.append(modules_package)
+                    break
             except KeyError:
                 pass
 
@@ -826,20 +1031,12 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
 
 def nvidia_desktop_pre_installation_hook(to_install):
     '''Applies changes that need to happen before installing the NVIDIA drivers'''
-    with_nvidia_kms = False
-
     # Enable KMS if nvidia >= 470
     for package_name in to_install:
-        if package_name.startswith('nvidia-driver-'):
-            try:
-                version = int(package_name.split('-')[-1])
-            except ValueError:
-                pass
-            finally:
-                with_nvidia_kms = version >= 470
-
-    if with_nvidia_kms:
-        set_nvidia_kms(1)
+        nvidia_info = NvidiaPkgNameInfo(package_name)
+        if nvidia_info.get_major_version() >= 470:
+            set_nvidia_kms(1)
+            return
 
 
 def nvidia_desktop_post_installation_hook():
@@ -895,7 +1092,7 @@ def _process_driver_string(string):
     return driver
 
 
-def gpgpu_install_filter(packages, drivers_str):
+def gpgpu_install_filter(packages, drivers_str, get_recommended=True):
     drivers = []
     allow = []
     result = {}
@@ -988,14 +1185,17 @@ def gpgpu_install_filter(packages, drivers_str):
                     result[p] = packages[p]
                 else:
                     # print('before recommended: %s' % packages[p])
-                    if packages[p].get('recommended'):
+                    if get_recommended:
+                        if packages[p].get('recommended'):
+                            result[p] = packages[p]
+                    else:
                         result[p] = packages[p]
                         # print('Found "recommended" flavour in %s' % (packages[p]))
                 break
     return result
 
 
-def auto_install_filter(packages, drivers_str=''):
+def auto_install_filter(packages, drivers_str='', get_recommended=True):
     '''Get packages which are appropriate for automatic installation.
 
     Return the subset of the given list of packages which are appropriate for
@@ -1019,7 +1219,10 @@ def auto_install_filter(packages, drivers_str=''):
 
     result = {}
     for p in allow:
-        if 'recommended' not in packages[p] or packages[p]['recommended']:
+        if get_recommended:
+            if 'recommended' not in packages[p] or packages[p]['recommended']:
+                result[p] = packages[p]
+        else:
             result[p] = packages[p]
     return result
 
@@ -1093,59 +1296,140 @@ def _pkg_support_from_cache(x):
     return None
 
 
+def _pkg_open_preferred_from_cache(x):
+    '''Look up driver and return if open package is preferred'''
+    if lookup_cache.get(x):
+        return lookup_cache.get(x).get('open_preferred')
+    return False
+
+
+def _get_fit_level(x):
+    '''decide how well the package fits desktop environment'''
+
+    # for desktop non-server packages are prefered to -server ones
+    # if a package preferes open variants, these come on top
+
+    if _pkg_open_preferred_from_cache(x):
+        if re.search(r'(\d{3})-open$', x):
+            return 3
+
+        if re.search(r'(\d{3})$', x):
+            return 2
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-server$', x):
+            return 0
+    else:
+        if re.search(r'(\d{3})$', x):
+            return 3
+
+        if re.search(r'(\d{3})-server$', x):
+            return 2
+
+        if re.search(r'(\d{3})-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 0
+    return 0
+
+
+def _get_fit_level_gpgpu(x):
+    '''decide how well the package fits server environment'''
+
+    # for server environment -server packages are prefered to non-server ones
+    # if a package preferes open variants, these come on top
+
+    if _pkg_open_preferred_from_cache(x):
+        if re.search(r'(\d{3})-server-open$', x):
+            return 3
+
+        if re.search(r'(\d{3})-server$', x):
+            return 2
+
+        if re.search(r'(\d{3})-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})$', x):
+            return 0
+    else:
+        if re.search(r'(\d{3})-server$', x):
+            return 3
+
+        if re.search(r'(\d{3})$', x):
+            return 2
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-open$', x):
+            return 0
+    return 0
+
+
 def _cmp_gfx_alternatives(x, y):
     '''Compare two graphics driver names in terms of preference. (desktop)
 
+    -open by default sorts after non-open, unless a driver preferes open variant
     -server always sorts after non-server.
     LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
     Legacy always sorts before Beta.
     '''
-    if x.endswith('-server') and not y.endswith('-server'):
-        return -1
-    if not x.endswith('-server') and y.endswith('-server'):
-        return 1
 
-    preferred_support = ['PB', 'LTSB']
-
-    x_score = 0
-    y_score = 0
-
-    x_support = _pkg_support_from_cache(x)
-    y_support = _pkg_support_from_cache(y)
-
-    if x_support in preferred_support:
-        x_score += 100
-
-    if y_support in preferred_support:
-        y_score += 100
-
-    if x > y:
-        x_score += 1
-    elif x < y:
-        y_score += 1
-
-    if ((x_score >= 100) or (y_score >= 100)):
-        if x_score > y_score:
+    if x.startswith('nvidia-') and y.startswith('nvidia-'):
+        if _get_fit_level(x) > _get_fit_level(y):
             return 1
-        elif x_score < y_score:
+        if _get_fit_level(x) < _get_fit_level(y):
             return -1
 
-    if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
-        return 1
-    if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
-        return -1
-    if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
-        return 1
-    if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
-        return -1
-    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
-        return -1
-    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
-        return 1
-    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
-        return -1
-    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
-        return 1
+        preferred_support = ['PB', 'LTSB']
+
+        x_score = 0
+        y_score = 0
+
+        x_support = _pkg_support_from_cache(x)
+        y_support = _pkg_support_from_cache(y)
+
+        if x_support in preferred_support:
+            x_score += 100
+
+        if y_support in preferred_support:
+            y_score += 100
+
+        if x > y:
+            x_score += 1
+        elif x < y:
+            y_score += 1
+
+        if ((x_score >= 100) or (y_score >= 100)):
+            if x_score > y_score:
+                return 1
+            elif x_score < y_score:
+                return -1
+
+        if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
+            return 1
+        if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
+            return -1
+        if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
+            return 1
+        if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
+            return -1
+        if _pkg_support_from_cache(x) == 'NFB' and _pkg_support_from_cache(y) != 'NFB':
+            return -1
+        if _pkg_support_from_cache(x) != 'NFB' and _pkg_support_from_cache(y) == 'NFB':
+            return 1
+        if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
+            return -1
+        if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
+            return 1
+        if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
+            return -1
+        if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
+            return 1
+
     if x < y:
         return -1
     if x > y:
@@ -1157,31 +1441,40 @@ def _cmp_gfx_alternatives(x, y):
 def _cmp_gfx_alternatives_gpgpu(x, y):
     '''Compare two graphics driver names in terms of preference. (server)
 
+    -open by default sorts after non-open, unless a driver preferes open variant
     -server always sorts before non-server.
+    non server package always sorts after server package
     LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
     Legacy always sorts before Beta.
     '''
-    if x.endswith('-server') and not y.endswith('-server'):
-        return 1
-    if not x.endswith('-server') and y.endswith('-server'):
-        return -1
 
-    if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
-        return 1
-    if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
-        return -1
-    if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
-        return 1
-    if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
-        return -1
-    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
-        return -1
-    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
-        return 1
-    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
-        return -1
-    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
-        return 1
+    if x.startswith('nvidia-') and y.startswith('nvidia-'):
+        if _get_fit_level_gpgpu(x) > _get_fit_level_gpgpu(y):
+            return 1
+        if _get_fit_level_gpgpu(x) < _get_fit_level_gpgpu(y):
+            return -1
+
+        if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
+            return 1
+        if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
+            return -1
+        if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
+            return 1
+        if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
+            return -1
+        if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
+            return -1
+        if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
+            return 1
+        if _pkg_support_from_cache(x) == 'NFB' and _pkg_support_from_cache(y) != 'NFB':
+            return -1
+        if _pkg_support_from_cache(x) != 'NFB' and _pkg_support_from_cache(y) == 'NFB':
+            return 1
+        if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
+            return -1
+        if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
+            return 1
+
     if x < y:
         return -1
     if x > y:
@@ -1277,14 +1570,13 @@ def get_linux_modules_metapackage(apt_cache, candidate):
 
     depcache = apt_pkg.DepCache(apt_cache)
 
-    if 'nvidia' not in candidate:
-        logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping', candidate)
+    nvidia_info = NvidiaPkgNameInfo(candidate)
+    if not nvidia_info.is_valid:
+        logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping' % candidate)
         return metapackage
 
-    pattern = re.compile('nvidia-([0-9]+)')
-    match = pattern.match(candidate)
-    if match:
-        logging.debug('Legacy driver detected: %s. Skipping.', candidate)
+    if nvidia_info.has_obsolete_name_scheme():
+        logging.debug('Legacy driver detected: %s. Skipping.' % candidate)
         return metapackage
 
     linux_image_meta = get_linux_image(apt_cache)
@@ -1294,16 +1586,10 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     if linux_image:
         linux_flavour = linux_image.replace('linux-image-', '')
     else:
-        logging.debug('No linux-image can be found for %s. Skipping.', candidate)
+        logging.debug('No linux-image can be found for %s. Skipping.' % candidate)
         return metapackage
 
-    pattern = re.compile('nvidia-.*-([0-9]+)(.*)')
-    match = pattern.match(candidate)
-    if not match:
-        logging.debug('No flavour can be found in %s. Skipping.', candidate)
-        return metapackage
-
-    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
+    candidate_flavour = nvidia_info.get_flavour()
     linux_modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_flavour)
 
     try:
@@ -1338,8 +1624,11 @@ def get_linux_modules_metapackage(apt_cache, candidate):
                         if dep.parent_pkg.name.startswith('linux-modules-nvidia-')]
 
         pick = ''
-        modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour,
-                                                            get_linux_image(apt_cache).replace('linux-image-', ''))
+        candidate_suffix = get_linux_image(apt_cache).replace('linux-image-', '')
+        if candidate_suffix == "virtual":
+            candidate_suffix = "generic"
+        modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, candidate_suffix)
+
         for dep in reverse_deps:
             if dep == modules_candidate:
                 pick = dep
